@@ -1,12 +1,12 @@
-;; Ada mode compiling functionality provided by the 'gnat'
-;; tool.
+;; ada-gnat-compile.el --- Ada mode compiling functionality provided by 'gnat'  -*- lexical-binding:t -*-
+;; Includes related functions, such as gnatprep support.
 ;;
 ;; These tools are all Ada-specific; use Makefiles for multi-language
 ;; GNAT compilation tools.
 ;;
 ;; GNAT is provided by AdaCore; see http://libre.adacore.com/
 ;;
-;;; Copyright (C) 2012, 2013  Free Software Foundation, Inc.
+;;; Copyright (C) 2012 - 2017  Free Software Foundation, Inc.
 ;;
 ;; Author: Stephen Leake <stephen_leake@member.fsf.org>
 ;; Maintainer: Stephen Leake <stephen_leake@member.fsf.org>
@@ -34,8 +34,10 @@
 ;; By default, ada-mode is configured to load this file, so nothing
 ;; special needs to done to use it.
 
+(require 'cl-lib)
 (require 'compile)
 (require 'gnat-core)
+(require 'ada-fix-error)
 
 ;;;;; code
 
@@ -51,63 +53,63 @@ For `compilation-filter-hook'."
     ;; `compilation-mode-font-lock-keywords'.
     ;;
     ;; compilation-filter might insert partial lines, or it might insert multiple lines
-    (when (bolp)
-      (while (not (eobp))
-	;; We don't want 'next-error' to always go to secondary
-	;; references, so we _don't_ set 'compilation-message text
-	;; property. Instead, we set 'ada-secondary-error, so
-	;; `ada-goto-secondary-error' will handle it. We also set
-	;; fonts, so the user can see the reference.
+    (goto-char (line-beginning-position))
+    (while (not (eobp))
+      ;; We don't want 'next-error' to always go to secondary
+      ;; references, so we _don't_ set 'compilation-message text
+      ;; property. Instead, we set 'ada-secondary-error, so
+      ;; `ada-goto-secondary-error' will handle it. We also set
+      ;; fonts, so the user can see the reference.
 
-	;; typical secondary references look like:
-	;;
-	;; trivial_productions_test.adb:57:77:   ==> in call to "Get" at \
-	;;    opentoken-token-enumerated-analyzer.ads:88, instance at line 41
-	;;
-	;; c:/foo/bar/lookahead_test.adb:379:14: found type access to "Standard.String" defined at line 379
-	;;
-	;; lookahead_test.ads:23:09: "Name" has been inherited from subprogram at aunit-simple_test_cases.ads:47
-	;;
-	;; lalr.adb:668:37: non-visible declaration at analyzer.ads:60, instance at parser.ads:38
-	;;
-	;; save the file from the primary reference, look for "*.ad?:nn", "at line nnn"
+      ;; typical secondary references look like:
+      ;;
+      ;; trivial_productions_test.adb:57:77:   ==> in call to "Get" at \
+      ;;    opentoken-token-enumerated-analyzer.ads:88, instance at line 41
+      ;;
+      ;; c:/foo/bar/lookahead_test.adb:379:14: found type access to "Standard.String" defined at line 379
+      ;;
+      ;; lookahead_test.ads:23:09: "Name" has been inherited from subprogram at aunit-simple_test_cases.ads:47
+      ;;
+      ;; lalr.adb:668:37: non-visible declaration at analyzer.ads:60, instance at parser.ads:38
+      ;;
+      ;; save the file from the primary reference, look for "*.ad?:nn", "at line nnn"
 
-	(let (file)
-	  (when (looking-at "^\\(\\(.:\\)?[^ :\n]+\\):")
-	    (setq file (match-string-no-properties 1)))
+      (let (file)
+	(when (looking-at "^\\(\\(.:\\)?[^ :\n]+\\):")
+	  (setq file (match-string-no-properties 1)))
 
-	  (skip-syntax-forward "^-"); space following primary reference
+	(skip-syntax-forward "^-"); space following primary reference
 
-	  (while (search-forward-regexp "\\s-\\(\\([^[:blank:]]+\\.[[:alpha:]]+\\):\\([0-9]+\\)\\)"
-					(line-end-position) t)
+	(while (search-forward-regexp "\\s-\\(\\([^[:blank:]]+\\.[[:alpha:]]+\\):\\([0-9]+\\)\\)"
+				      (line-end-position) t)
 
-	    (goto-char (match-end 0))
-	    (with-silent-modifications
-	      (compilation--put-prop 2 'font-lock-face compilation-info-face); file
-	      (compilation--put-prop 3 'font-lock-face compilation-line-face); line
-	      (put-text-property
-	       (match-beginning 0) (match-end 0)
-	       'ada-secondary-error
-	       (list
-		(match-string-no-properties 2); file
-		(string-to-number (match-string-no-properties 3)); line
-		1)); column
-	      ))
+	  (goto-char (match-end 0))
+	  (with-silent-modifications
+	    (compilation--put-prop 2 'font-lock-face compilation-info-face); file
+	    (compilation--put-prop 3 'font-lock-face compilation-line-face); line
+	    (put-text-property
+	     (match-beginning 0) (match-end 0)
+	     'ada-secondary-error
+	     (list
+	      (match-string-no-properties 2); file
+	      (string-to-number (match-string-no-properties 3)); line
+	      0)); Emacs column; zero indexed
+	    ))
 
-	  (when (search-forward-regexp "\\(at line \\)\\([0-9]+\\)" (line-end-position) t)
-	    (with-silent-modifications
-	      (compilation--put-prop 1 'font-lock-face compilation-info-face); "at line" instead of file
-	      (compilation--put-prop 2 'font-lock-face compilation-line-face); line
-	      (put-text-property
-	       (match-beginning 1) (match-end 1)
-	       'ada-secondary-error
-	       (list
-		file
-		(string-to-number (match-string-no-properties 2)); line
-		1)); column
-	      ))
-	  (forward-line 1))
-	))
+	(when (search-forward-regexp "\\(at line \\)\\([0-9]+\\)" (line-end-position) t)
+	  (with-silent-modifications
+	    (compilation--put-prop 1 'font-lock-face compilation-info-face); "at line" instead of file
+	    (compilation--put-prop 2 'font-lock-face compilation-line-face); line
+	    (put-text-property
+	     (match-beginning 1) (match-end 1)
+	     'ada-secondary-error
+	     (list
+	      file
+	      (string-to-number (match-string-no-properties 2)); line
+	      1)); column
+	    ))
+	(forward-line 1))
+      )
     ))
 
 (defun ada-gnat-debug-filter ()
@@ -165,7 +167,7 @@ Prompt user if more than one."
       (completing-read "correct spelling: " choices))
      )))
 
-(defun ada-gnat-fix-error (msg source-buffer source-window)
+(defun ada-gnat-fix-error (_msg source-buffer _source-window)
   "For `ada-gnat-fix-error-hook'."
   (let ((start-pos (point))
 	message-column
@@ -192,45 +194,50 @@ Prompt user if more than one."
 	  ;; Then style errors.
 
 	  ((looking-at (concat ada-gnat-quoted-name-regexp " is not visible"))
-	   (let ((ident (match-string 1))
-		 (done nil)
+	   (let ((done nil)
 		 (file-line-struct (progn (beginning-of-line) (ada-get-compilation-message)))
 		 pos choices unit-name)
 	     ;; next line may contain a reference to where ident is
 	     ;; defined; if present, it will have been marked by
-	     ;; ada-gnat-compilation-filter
+	     ;; ada-gnat-compilation-filter:
+	     ;;
+	     ;; gnatquery.adb:255:13: "Has_Element" is not visible
+	     ;; gnatquery.adb:255:13: non-visible declaration at a-convec.ads:68, instance at gnatcoll-arg_lists.ads:157
+	     ;; gnatquery.adb:255:13: non-visible declaration at a-coorse.ads:62, instance at gnatcoll-xref.ads:912
+	     ;; gnatquery.adb:255:13: non-visible declaration at a-coorse.ads:62, instance at gnatcoll-xref.ads:799
+	     ;; gnatquery.adb:255:13: non-visible declaration at gnatcoll-xref.ads:314
 	     ;;
 	     ;; or the next line may contain "multiple use clauses cause hiding"
 	     ;;
 	     ;; the lines after that may contain alternate matches;
 	     ;; collect all, let user choose.
-	     (while (not done)
-	       (forward-line 1)
-	       (unless (looking-at ".* multiple use clauses cause hiding")
-		 (setq done (not
-			     (and
-			      (equal file-line-struct (ada-get-compilation-message))
-			      (let ((limit (1- (line-end-position))))
-				;; 1- because next compilation error is at next line beginning
+	     (forward-line 1)
+	     (unless (looking-at ".* multiple use clauses cause hiding")
+	       (while (not done)
+		 (let ((limit (1- (line-end-position))))
+		   ;; 1- because next compilation error is at next line beginning
+		   (setq done (not
+			       (and
+				(equal file-line-struct (ada-get-compilation-message))
 				(setq pos (next-single-property-change (point) 'ada-secondary-error nil limit))
-				(< pos limit)))))
-		 (when (not done)
-		   (let* ((item (get-text-property pos 'ada-secondary-error))
-			  (unit-file (nth 0 item)))
-		     (add-to-list 'choices (ada-ada-name-from-file-name unit-file)))))
-	       );; while
+				(< pos limit))))
+		   (when (not done)
+		     (let* ((item (get-text-property pos 'ada-secondary-error))
+			    (unit-file (nth 0 item))
+                            (choice (ada-ada-name-from-file-name unit-file)))
+                       (unless (member choice choices) (push choice choices))
+		       (goto-char (1+ pos))
+		       (goto-char (1+ (next-single-property-change (point) 'ada-secondary-error nil limit)))
+		       (when (eolp) (forward-line 1))
+		       ))
+		   )));; unless while let
 
-	     (cond
-	      ((= 0 (length choices))
-	       (setq unit-name nil))
-
-	      ((= 1 (length choices))
-	       (setq unit-name (car choices)))
-
-	      (t ;; multiple choices
-	       (setq unit-name
-		     (completing-read "package name: " choices)))
-	      );; cond
+	     (setq unit-name
+		   (cond
+		    ((= 0 (length choices)) nil)
+		    ((= 1 (length choices)) (car choices))
+		    (t ;; multiple choices
+		     (completing-read "package name: " choices))))
 
 	     (when unit-name
 	       (pop-to-buffer source-buffer)
@@ -299,7 +306,13 @@ Prompt user if more than one."
 	   t)
 
 ;;;; strings
-	  ((looking-at (concat "misspelling of " ada-gnat-quoted-name-regexp))
+	  ((looking-at "package \"Ada\" is hidden")
+	   (pop-to-buffer source-buffer)
+	   (forward-word -1)
+	   (insert "Standard.")
+	   t)
+
+	  ((looking-at (concat "\\(?:possible \\)?misspelling of " ada-gnat-quoted-name-regexp))
 	   (let ((expected-name (match-string 1)))
 	     (pop-to-buffer source-buffer)
 	     (looking-at ada-name-regexp)
@@ -320,6 +333,19 @@ Prompt user if more than one."
 	     (insert expected-name))
 	   t)
 
+	  ((looking-at (concat "\"end loop " ada-name-regexp ";\" expected"))
+	   (let ((expected-name (match-string 1)))
+	     (pop-to-buffer source-buffer)
+	     (if (looking-at (concat "end loop " ada-name-regexp ";"))
+		 (progn
+		   (goto-char (match-end 1))   ; just before ';'
+		   (delete-region (match-beginning 1) (match-end 1)))
+	       ;; else we have just 'end loop;'
+	       (forward-word 2)
+	       (insert " "))
+	     (insert expected-name))
+	   t)
+
 	  ((looking-at "expected an access type")
 	   (progn
 	     (set-buffer source-buffer)
@@ -329,16 +355,23 @@ Prompt user if more than one."
 	       t)))
 
 	  ((looking-at (concat "expected \\(private \\)?type " ada-gnat-quoted-name-regexp))
-	   (let ((type (match-string 2)))
-	     (forward-line 1)
-	     (move-to-column message-column)
-	     (when (or (looking-at "found type access")
-		       (looking-at "found type .*_Access_Type"))
-	       ;; assume just need '.all'
-	       (pop-to-buffer source-buffer)
+	   (forward-line 1)
+	   (move-to-column message-column)
+	   (cond
+	    ((looking-at "found type access")
+	     (pop-to-buffer source-buffer)
+	     (if (looking-at "'Access")
+		 (kill-word 1)
 	       (forward-word 1)
-	       (insert ".all")
-	       t)))
+	       (insert ".all"))
+	     t)
+	    ((looking-at "found type .*_Access_Type")
+	     ;; assume just need '.all'
+	     (pop-to-buffer source-buffer)
+	     (forward-word 1)
+	     (insert ".all")
+	     t)
+	    ))
 
 	  ((looking-at "extra \".\" ignored")
 	   (set-buffer source-buffer)
@@ -355,7 +388,9 @@ Prompt user if more than one."
 	   ;; also 'possible missing "with Ada.Text_IO; use Ada.Text_IO"' - ignoring the 'use'
 	   (let ((package-name (match-string-no-properties 1)))
 	     (pop-to-buffer source-buffer)
-	     ;; FIXME (later): should check if prefix is already with'd, extend it
+	     ;; Could check if prefix is already with'd, extend
+	     ;; it. But no one has reported that case yet; this
+	     ;; message only occurs for predefined Ada packages.
 	     (ada-fix-add-with-clause package-name))
 	   t)
 
@@ -389,10 +424,8 @@ Prompt user if more than one."
 	  ((looking-at (concat "operator for \\(private \\)?type " ada-gnat-quoted-name-regexp))
 	   (let ((type (match-string 2)))
 	     (pop-to-buffer source-buffer)
-	     (ada-goto-declarative-region-start)
-	     (newline-and-indent)
-	     (insert "use type " type ";"))
-	   t)
+	     (ada-fix-add-use-type type)
+	   t))
 
 	  ((looking-at "parentheses required for unary minus")
 	   (set-buffer source-buffer)
@@ -443,8 +476,7 @@ Prompt user if more than one."
 	   t)
 
 	  ((looking-at (concat "warning: formal parameter " ada-gnat-quoted-name-regexp " is not modified"))
-	   (let ((param (match-string 1))
-		 (mode-regexp "\"\\([in out]+\\)\"")
+	   (let ((mode-regexp "\"\\([in out]+\\)\"")
 		 new-mode
 		 old-mode)
 	     (forward-line 1)
@@ -472,7 +504,8 @@ Prompt user if more than one."
 	  ((looking-at (concat "warning: variable " ada-gnat-quoted-name-regexp " is assigned but never read"))
 	   (let ((param (match-string 1)))
 	     (pop-to-buffer source-buffer)
-	     (ada-goto-end)
+	     (ada-goto-end) ;; leaves point before semicolon
+	     (forward-char 1)
 	     (newline-and-indent)
 	     (insert "pragma Unreferenced (" param ");"))
 	   t)
@@ -486,17 +519,15 @@ Prompt user if more than one."
 
 ;;;; style errors
 	  ((looking-at "(style) \".*\" in wrong column")
-	   (progn
-	     (set-buffer source-buffer)
-	     (funcall indent-line-function))
+	   (set-buffer source-buffer)
+	   (funcall indent-line-function)
 	   t)
 
 	  ((looking-at "(style) bad capitalization, mixed case required")
-	   (progn
-	     (set-buffer source-buffer)
-	     (forward-word)
-	     (ada-case-adjust-identifier)
-	     t))
+	   (set-buffer source-buffer)
+	   (forward-word)
+	   (ada-case-adjust-identifier)
+	   t)
 
 	  ((looking-at (concat "(style) bad casing of " ada-gnat-quoted-name-regexp))
 	   (let ((correct (match-string-no-properties 1))
@@ -518,6 +549,18 @@ Prompt user if more than one."
 	   (funcall indent-line-function)
 	   t)
 
+	  ((looking-at "(style) \"exit \\(.*\\)\" required")
+	   (let ((name (match-string-no-properties 1)))
+	     (set-buffer source-buffer)
+	     (forward-word 1)
+	     (insert (concat " " name))
+	   t))
+
+	  ((looking-at "(style) misplaced \"then\"")
+	   (set-buffer source-buffer)
+	   (delete-indentation)
+	   t)
+
          ((looking-at "(style) missing \"overriding\" indicator")
           (set-buffer source-buffer)
           (cond
@@ -527,6 +570,11 @@ Prompt user if more than one."
            (t
             nil)))
 
+	  ((looking-at "(style) reserved words must be all lower case")
+	   (set-buffer source-buffer)
+	   (downcase-word 1)
+	   t)
+
 	  ((looking-at "(style) space not allowed")
 	   (set-buffer source-buffer)
 	   ;; Error places point on space. More than one trailing space
@@ -550,7 +598,24 @@ Prompt user if more than one."
 
 (defun ada-gnat-compile-select-prj ()
   (setq ada-fix-error-hook 'ada-gnat-fix-error-hook)
+  (setq ada-prj-show-prj-path 'gnat-prj-show-prj-path)
   (add-to-list 'completion-ignored-extensions ".ali") ;; gnat library files
+  (add-hook 'ada-syntax-propertize-hook 'ada-gnat-syntax-propertize)
+  (add-hook 'ada-syntax-propertize-hook 'gnatprep-syntax-propertize)
+  (syntax-ppss-flush-cache (point-min));; force re-evaluate with hook.
+
+  ;; There is no common convention for a file extension for gnatprep files.
+  ;;
+  ;; find error locations in .gpr files
+  (setq compilation-search-path (append compilation-search-path (ada-prj-get 'prj_dir)))
+  (setq compilation-environment
+	(list
+	 (let ((process-environment (cl-copy-list (ada-prj-get 'proc_env))))
+	   (concat "GPR_PROJECT_PATH=" (getenv "GPR_PROJECT_PATH")))))
+
+  ;; must be after indentation engine setup, because that resets the
+  ;; indent function list.
+  (add-hook 'ada-mode-hook 'gnatprep-setup t)
 
   (add-hook 'compilation-filter-hook 'ada-gnat-compilation-filter)
 
@@ -562,12 +627,21 @@ Prompt user if more than one."
 (defun ada-gnat-compile-deselect-prj ()
   (setq ada-fix-error-hook nil)
   (setq completion-ignored-extensions (delete ".ali" completion-ignored-extensions))
+  (setq ada-syntax-propertize-hook (delq 'gnatprep-syntax-propertize ada-syntax-propertize-hook))
+  (setq ada-syntax-propertize-hook (delq 'ada-gnat-syntax-propertize ada-syntax-propertize-hook))
+  (syntax-ppss-flush-cache (point-min));; force re-evaluate with hook.
+
+  ;; don't need to delete from compilation-search-path; completely rewritten in ada-select-prj-file
+  (setq compilation-environment nil)
+
+  (setq ada-mode-hook (delq 'gnatprep-setup ada-mode-hook))
+
   (setq compilation-filter-hook (delete 'ada-gnat-compilation-filter compilation-filter-hook))
   (setq compilation-error-regexp-alist (delete 'gnat compilation-error-regexp-alist))
   )
 
 (defun ada-gnat-compile ()
-  "Set Ada mode global vars to use 'gnat' for compiling."
+  "Set Ada mode global vars to use `gnat' for compiling."
   (add-to-list 'ada-prj-file-ext-extra     "gpr")
   (add-to-list 'ada-prj-parser-alist       '("gpr" . gnat-parse-gpr))
   (add-to-list 'ada-select-prj-compiler    '(gnat  . ada-gnat-compile-select-prj))
@@ -576,6 +650,10 @@ Prompt user if more than one."
   (add-to-list 'ada-prj-parse-one-compiler   (cons 'gnat 'gnat-prj-parse-emacs-one))
   (add-to-list 'ada-prj-parse-final-compiler (cons 'gnat 'gnat-prj-parse-emacs-final))
 
+  (font-lock-add-keywords 'ada-mode
+   ;; gnatprep preprocessor line
+   (list (list "^[ \t]*\\(#.*\n\\)"  '(1 font-lock-preprocessor-face t))))
+
   (add-hook 'ada-gnat-fix-error-hook 'ada-gnat-fix-error))
 
 (provide 'ada-gnat-compile)
@@ -597,7 +675,7 @@ Prompt user if more than one."
    ;;   foo.c:2 : `TRUE' undeclared here (not in a function)
    "^\\(\\(.:\\)?[^ :\n]+\\):\\([0-9]+\\)\\s-?:?\\([0-9]+\\)?" 1 3 4))
 
-(unless (default-value ada-compiler)
+(unless (default-value 'ada-compiler)
     (set-default 'ada-compiler 'gnat))
 
 ;; end of file
